import sys
import os, errno, glob
import dircache, shutil
import re
from plistlib import Plist
from distutils import dir_util, file_util

from project import *
import utils

class Bundler:
    def __init__(self, project):
        self.project = project

        self.project_dir = project.get_project_dir()

        plist_path = self.project.get_plist_path()
        self.plist = Plist.fromFile(plist_path)

        # List of paths that should be recursively searched for
        # binaries that are used to find library dependencies.
        self.binary_paths = []
        #List of frameworks moved into the bundle which need to be set
        #up for private use.
        self.frameworks = []

        # Create the bundle in a temporary location first and move it
        # to the final destination when done.
        self.meta = project.get_meta()
        self.bundle_path = os.path.join(self.meta.dest, "." + project.get_bundle_name() + ".app")

    def recursive_rm(self, dirname):
        # Extra safety ;)
        if dirname in [ "/", os.getenv("HOME"), os.path.join(os.getenv("HOME"), "Desktop"), self.meta.dest ]:
            print "Eek, trying to remove a bit much, eh? (%s)" % (dirname)
            sys.exit(1)

        if not os.path.exists(dirname):
            return
        
        files = dircache.listdir(dirname)
        for file in files:
            path = os.path.join (dirname, file)
            if os.path.isdir(path):
                self.recursive_rm(path)
            else:
                retval = os.unlink(path)
        if (os.path.islink(dirname)):
            os.unlink(dirname)
        else:
            os.rmdir(dirname)

    def create_skeleton(self):
        utils.makedirs(self.project.get_bundle_path("Contents/Resources"))
        utils.makedirs(self.project.get_bundle_path("Contents/MacOS"))

    def create_pkglist(self):
        path = self.project.get_bundle_path("Contents", "PkgInfo")
        path = self.project.evaluate_path(path)
        f = open (path, "w")
        f.write(self.plist.CFBundlePackageType)
        f.write(self.plist.CFBundleSignature)
        f.close()

    def copy_plist(self):
        path = Path(self.project.get_plist_path(),
                    self.project.get_bundle_path("Contents/Info.plist"))
        self.copy_path(path)

    def create_pango_setup(self):
        # Create a temporary pangorc file just for creating the
        # modules file with the right modules.
        module_version = utils.evaluate_pkgconfig_variables("${pkg:pango:pango_module_version}")
        modulespath = self.project.get_bundle_path("Contents/Resources/lib/pango/" +
                                                   module_version +
                                                   "modules/")

        from distutils.version import StrictVersion as V
        import tempfile

        fd, tmp_filename = tempfile.mkstemp()
        f = os.fdopen(fd, "w")
        f.write("[Pango]\n")
        f.write("ModulesPath=" + modulespath)
        f.write("\n")
        f.close()

        cmd = "PANGO_RC_FILE=" + tmp_filename + " pango-querymodules"
        f = os.popen(cmd)

        path = self.project.get_bundle_path("Contents/Resources/etc/pango")
        utils.makedirs(path)
        fout = open(os.path.join(path, "pango.modules"), "w")

        if V(module_version) < V('1.8.0'):
            prefix_path  = os.path.join("Contents", "Resources")
        else:
            prefix_path = os.path.join("Contents", "Resources", "lib", "pango",
                                       module_version, "modules/")

        prefix = self.project.get_bundle_path(prefix_path)

        for line in f:
            line = line.strip()
            if line.startswith("# ModulesPath"):
                continue

            # Replace the hardcoded bundle path with @executable_path...
            if line.startswith(prefix):
                line = line[len(prefix):]
#Newer versions of pango have been modified to look in the right place
#for modules (providing the PANGO_LIB_DIR is set to point to the
#bundle_lib folder).
                if V(module_version) < V('1.8.0'):
                    line = "@executable_path/../Resources" + line

            fout.write(line)
            fout.write("\n")
        fout.close()

        os.unlink(tmp_filename)

        # Create the final pangorc file
        path = self.project.get_bundle_path("Contents/Resources/etc/pango")
        utils.makedirs(path)
        f = open(os.path.join(path, "pangorc"), "w")
        f.write("[Pango]\n")
#Pango 2.32 (corresponding to module_version 1.8.0) and later don't
#interpret "./" to mean the same directory that the rc file is in, so
#this doesn't work any more. However, pango does know to look in the
#bundle directory (when given the right environment variable), so we
#don't need this, either.
        if V(module_version) < V('1.8.0'):
            f.write("ModuleFiles=./pango.modules\n")
        f.close()

    def create_gtk_immodules_setup(self):
        path = self.project.get_bundle_path("Contents/Resources")
        cmd = "GTK_EXE_PREFIX=" + path + " gtk-query-immodules-" + self.project.get_gtk_version()
        f = os.popen(cmd)

        path = self.project.get_bundle_path("Contents/Resources/etc/", 
                                            self.project.get_gtk_dir())
        utils.makedirs(path)
        fout = open(os.path.join(path, "gtk.immodules"), "w")

        prefix = "\"" + self.project.get_bundle_path("Contents/Resources")

        for line in f:
            line = line.strip()
            if line.startswith("#"):
                continue

            # Replace the hardcoded bundle path with @executable_path...
            if line.startswith(prefix):
                line = line[len(prefix):]
                line = "\"@executable_path/../Resources" + line
            fout.write(line)
            fout.write("\n")
        fout.close()

    def create_gdk_pixbuf_loaders_setup(self):
        modulespath = ""
        if os.path.exists(os.path.join(self.project.get_prefix(), "lib", 
                                       "gdk-pixbuf-2.0")):

            modulespath = self.project.get_bundle_path("Contents/Resources/lib/",
                                                     "gdk-pixbuf-2.0", 
                                                     "${pkg:gdk-pixbuf-2.0:gdk_pixbuf_binary_version}",
                                                     "loaders")
        elif os.path.exists(os.path.join(self.project.get_prefix(), "lib", 
                                       "gdk-pixbuf-3.0")):
            modulespath = self.project.get_bundle_path("Contents/Resources/lib/",
                                                     "gdk-pixbuf-3.0", 
                                                     "${pkg:gdk-pixbuf-3.0:gdk_pixbuf_binary_version}",
                                                     "loaders")
        else:
            modulespath = self.project.get_bundle_path("Contents/Resources/lib/",
                                                   self.project.get_gtk_dir(),
                                                   "${pkg:" + self.meta.gtk + ":gtk_binary_version}",
                                                   "loaders")
        modulespath = utils.evaluate_pkgconfig_variables (modulespath)
        cmd = "GDK_PIXBUF_MODULEDIR=" + modulespath + " gdk-pixbuf-query-loaders"
        f = os.popen(cmd)

        path = self.project.get_bundle_path("Contents/Resources/etc/", 
                                            self.project.get_gtk_dir())
        utils.makedirs(path)
        fout = open(os.path.join(path, "gdk-pixbuf.loaders"), "w")

        prefix = "\"" + self.project.get_bundle_path("Contents/Resources")
        for line in f:
            line = line.strip()
            if line.startswith("#"):
                continue

            # Replace the hardcoded bundle path with @executable_path...
            if line.startswith(prefix):
                line = line[len(prefix):]
                line = "\"@executable_path/../Resources" + line
            fout.write(line)
            fout.write("\n")
        fout.close()

    def copy_binaries(self, binaries):
        for path in binaries:
            dest = self.copy_path(path)

            self.binary_paths.append(dest)

            # Clean up duplicates
            self.binary_paths = list(set(self.binary_paths))

            # Clean out any libtool (*.la) files and static libraries
            if os.path.isdir(dest):
                for root, dirs, files in os.walk(dest):
                    for name in filter(lambda l: l.endswith(".la") or l.endswith(".a"), files):
                        os.remove(os.path.join(root, name))

    # Copies from Path.source to Path.dest, evaluating any variables
    # in the paths, and returns the real dest.
    def copy_path(self, Path):
        _doRecurse = False
        source = self.project.evaluate_path(Path.source)
        if Path.dest:
            dest = self.project.evaluate_path(Path.dest)
        else:
            # Source must begin with a prefix if we don't have a
            # dest. Skip past the source prefix and replace it with
            # the right bundle path instead.
            p = re.compile("^\${prefix(:.*?)?}/")
            m = p.match(Path.source)
            if m:
                relative_dest = self.project.evaluate_path(Path.source[m.end():])
                dest = self.project.get_bundle_path("Contents/Resources", relative_dest)
            else:
                print "Invalid bundle file, missing or invalid 'dest' property: " + Path.dest
                sys.exit(1)

        (dest_parent, dest_tail) = os.path.split(dest)
        utils.makedirs(dest_parent)

        # Check that the source only has wildcards in the last component.
        p = re.compile("[\*\?]")
        (source_parent, source_tail) = os.path.split(source)
        if p.search(source_parent):
            print "Can't have wildcards except in the last path component: " + source
            sys.exit(1)

        if p.search(source_tail):
            source_check = source_parent
            if Path.recurse:
                _doRecurse = True
        else:
            source_check = source
        if not os.path.exists(source_check):
            print "Cannot find source to copy: " + source
            sys.exit(1)

        # If the destination has a wildcard as last component (copied
        # from the source in dest-less paths), ignore the tail.
        if p.search(dest_tail):
            dest = dest_parent

        if _doRecurse:
            for root, dirs, files in os.walk(source_parent):
                destdir = os.path.join(dest,
                                       os.path.relpath(root, source_parent))
                utils.makedirs(destdir)
                for globbed_source in glob.glob(os.path.join(root,
                                                             source_tail)):
                    try:
#                        print "Copying %s to %s" % (globbed_source, destdir)
                        shutil.copy(globbed_source, destdir)
                    except EnvironmentError, e:
                        if e.errno == errno.ENOENT:
                            print "Warning, source file missing: " + globbed_source
                        elif e.errno == errno.EEXIST:
                            print "Warning, path already exits: " + dest
                        else:
                            print "Error %s when copying file: %s" % ( str(e), globbed_source )
                            sys.exit(1)

        else:
            for globbed_source in glob.glob(source):
                try:
                    if os.path.isdir(globbed_source):
                        #print "dir: %s => %s" % (globbed_source, dest)
                        dir_util.copy_tree (str(globbed_source), str(dest),
                                            preserve_mode=1,
                                            preserve_times=1,
                                            preserve_symlinks=1,
                                            update=1,
                                            verbose=1,
                                            dry_run=0)
                    else:
                        #print "file: %s => %s" % (globbed_source, dest)
                        file_util.copy_file (str(globbed_source), str(dest),
                                            preserve_mode=1,
                                            preserve_times=1,
                                            update=1,
                                            link=None,
                                            verbose=1,
                                            dry_run=0)
                except EnvironmentError, e:
                    if e.errno == errno.ENOENT:
                        print "Warning, source file missing: " + globbed_source
                    elif e.errno == errno.EEXIST:
                        print "Warning, path already exits: " + dest
                    else:
                        print "Error %s when copying file: %s" %( str(e), globbed_source )
                        sys.exit(1)
        return dest

    # Lists all the binaries copied in so far. Used in the library
    # dependency resolution and icon theme lookup.
    def list_copied_binaries(self):
        def filter_path(path):
            if os.path.islink(path):
                return False
            if path.endswith(".so") or path.endswith(".dylib") or os.access(path, os.X_OK):
                return True
            return False
        paths = []
        for path in self.binary_paths:
            if os.path.isdir(path):
                for root, dirs, files in os.walk(path):
                    paths.extend(map(lambda l: os.path.join(root, l), files))
            else:
                paths.append(path)

        paths = filter(filter_path, paths)
        return list(set(paths))

    def resolve_library_dependencies(self):
        # Get the libraries we link to, filter out anything that
        # doesn't come from any of the prefixes we have declared. Then
        # copy in the resulting list, and repeat until the list
        # doesn't change. At that point, all dependencies are
        # resolved.
        n_iterations = 0
        n_paths = 0
        paths = self.list_copied_binaries()
        while n_paths != len(paths):
            cmds = [ "otool -L " ]
            for path in paths:
                cmds.append(path + " ")

            cmd = ''.join(cmds)
            f = os.popen(cmd)

            prefixes = self.meta.prefixes

            def relative_path_map(line):
                if not os.path.isabs(line):
                    for prefix in prefixes.values():
                        path = os.path.join(prefix, "lib", line)
                        if os.path.exists(path):
                            return path
                    print "Cannot find a matching prefix for %s" % (line)
                return line

            def prefix_filter(line):
                if not "(compatibility" in line:
                    return False

                if line.startswith("/usr/X11"):
                    print "Warning, found X11 library dependency, you most likely don't want that:", line.strip().split()[0]

                if os.path.isabs(line):
                    for prefix in prefixes.values():
                        if prefix in line:
                            return True
                    
                    if not line.startswith("/usr/lib") and not line.startswith("/System/Library"):
                        print "Warning, library not available in any prefix:", line.strip().split()[0]

                    return False

                return True

            lines = filter(prefix_filter, [line.strip() for line in f])
            p = re.compile("(.*\.dylib\.?.*)\s\(compatibility.*$")
            lines = utils.filterlines(p, lines)
            lines = map(relative_path_map, lines)
#When you need to track down errors, uncomment this blocK
#            for path in paths:
#                cmd = "otool -L %s" % path
#                print path
#                f = os.popen(cmd)
#                lines = filter(prefix_filter, [line.strip() for line in f])

            new_libraries = []
            for library in set(lines):
                # Replace the real path with the right prefix so we can
                # create a Path object.
                for (key, value) in prefixes.items():
                    if library.startswith(value):
                        path = Path("${prefix:" + key + "}" + library[len(value):])
                        new_libraries.append(path)

            n_paths = len(paths)
            n_iterations += 1
            if n_iterations > 10:
                print "Too many tries to resolve library dependencies"
                sys.exit(1)
            
            self.copy_binaries(new_libraries)
            paths = self.list_copied_binaries()

    def run_install_name_tool(self):
        print "Running install name tool"

        paths = self.list_copied_binaries()
        prefixes = self.meta.prefixes

        # First change all references in libraries.
        for prefix in prefixes:
            prefix_path = self.project.get_prefix(prefix)
            print "Going through prefix: " + prefix_path
            for path in paths:
                cmd = os.path.join(os.path.dirname(__file__), "run-install-name-tool-change.sh") + " " + path + " " + prefix_path + " Resources" + " change"
                f = os.popen(cmd)
                for line in f:
                    print line

        # Then change the id of all libraries. Skipping this part for now
        #for path in paths:
        #    cmd = os.path.join(os.path.dirname(__file__), "run-install-name-tool-id.sh") + " " + path
        #    print cmd
        #    f = os.popen(cmd)
        #    for line in f:
        #        print line

        for framework in self.frameworks:
            fw_name, ext = os.path.splitext(os.path.basename(framework))
            fwl = os.path.join(framework, fw_name)
            print "Importing Framework: " + fwl
# Fix the framework IDs
            cmd = os.path.join(os.path.dirname(__file__), "run-install-name-tool-change.sh") + " " + fwl + " " + fw_name + " Frameworks" + " id"
            f = os.popen(cmd)
            for line in f:
                print line
# Fix the dependencies in other libraries
            for path in paths:
                cmd = os.path.join(os.path.dirname(__file__), "run-install-name-tool-change.sh") + " " + path + " " + fw_name + " Frameworks/" + fw_name + " change"
                f = os.popen(cmd)
                for line in f:
                    print line
#fix the dependencies in frameworks
            for ufw in self.frameworks:
                ufw_name, ext = os.path.splitext(os.path.basename(ufw))
                if ufw_name == fw_name:
                    continue
                ufwl = os.path.join(ufw, ufw_name)
                cmd = os.path.join(os.path.dirname(__file__), "run-install-name-tool-change.sh") + " " + ufwl + " " + fw_name + " Frameworks/" + fw_name + " change"
                f = os.popen(cmd)
                for line in f:
                    print line


    def strip_debugging(self):
        paths = self.list_copied_binaries()
        for path in paths:
            if path.endswith(".dylib") or path.endswith(".so"):
                os.chmod(path, 0644)
                os.system("strip -x " + path + " 2>/dev/null")
                os.chmod(path, 0444)
            else:
                os.chmod(path, 0755)
                os.system("strip -ur " + path + " 2>/dev/null")
                os.chmod(path, 0555)

#
# If you want to sign your application, set $APPLICATION_CERT with the
# appropriate certificate name in your default Keychain. This function
# will sign every binary in the bundle with the certificate and the
# bundle's id string.
#
    def sign_binaries(self):
        if not os.environ.has_key("APPLICATION_CERT"):
            return
        cert = os.getenv("APPLICATION_CERT")
        paths = self.list_copied_binaries()
        ident = self.project.get_bundle_id()
        for path in paths:
            cmdargs = ['codesign', '-s', cert, '-i', ident, path]
            result = os.spawnvp(os.P_WAIT, 'codesign', cmdargs)

            if result:
                raise OSError, '"' + " ".join(cmdargs) + '" failed %d' % result

    def copy_icon_themes(self):
        all_icons = set()

        themes = self.project.get_icon_themes()

        for theme in themes:
            self.copy_path(Path(os.path.join(theme.source, "index.theme")))

        for theme in themes:
            if theme.icons == IconTheme.ICONS_NONE:
                continue
            
            for root, dirs, files in os.walk(self.project.evaluate_path(theme.source)):
                for f in files:
                    (head, tail) = os.path.splitext(f)
                    if tail in [".png", ".svg"]:
                        all_icons.add(head)

        strings = set()

        # Get strings from binaries.
        for f in self.list_copied_binaries():
            p = os.popen("strings " + f)
            for string in p:
                string = string.strip()
                strings.add(string)

        # FIXME: Also get strings from glade files.

        used_icons = all_icons.intersection(strings)
        prefix = self.project.get_prefix()

        for theme in themes:
            if theme.icons == IconTheme.ICONS_NONE:
                continue

            for root, dirs, files in os.walk(self.project.evaluate_path(theme.source)):
                for f in files:
                    # Go through every file, if it matches the icon
                    # set, copy it.
                    (head, tail) = os.path.splitext(f)
                    if head in used_icons or theme.icons == IconTheme.ICONS_ALL:
                        path = os.path.join(root, f)

                        # Note: Skipping svgs for now, they are really
                        # big and not really used.
                        if path.endswith(".svg"):
                            continue

                        # Replace the real paths with the prefix macro
                        # so we can use copy_path.
                        self.copy_path(Path("${prefix}" + path[len(prefix):]))

        # Generate icon caches.
        for theme in themes:
            path = self.project.get_bundle_path("Contents/Resources/share/icons", theme.name)
            cmd = "gtk-update-icon-cache -f " + path + " 2>/dev/null"
            os.popen(cmd)

    def copy_translations(self):
        translations = self.project.get_translations()
        prefix = self.project.get_prefix()


        def name_filter(filename):
            path, fname = os.path.split(filename)
            name, ext = os.path.splitext(fname)
            if name != program.name:
                return False
            elif ext not in (".mo", ".po"):
                return False
            else:
                return True

        for program in translations:
            if program.name == "" or program.name == None:
                raise "No program name to tranlate!"

            source = self.project.evaluate_path(program.source)
            if source == None:
                raise "Failed to parse translation source!"
            for root, trees, files in os.walk(source):
                for file in filter(name_filter, files):
                    path = os.path.join(root, file)
                    self.copy_path(Path("${prefix}" + path[len(prefix):], 
                                        program.dest))


    def run(self):
        # Remove the temp location forcefully.
        path = self.project.evaluate_path(self.bundle_path)
        self.recursive_rm(path)

        final_path = os.path.join(self.meta.dest, self.project.get_bundle_name() + ".app")
        final_path = self.project.evaluate_path(final_path)

        if not self.meta.overwrite and os.path.exists(final_path):
            print "Bundle already exists: " + final_path
            sys.exit(1)

        self.create_skeleton()
        self.create_pkglist()
        self.copy_plist()

        # Note: could move this to xml file...
        self.copy_path(Path("${prefix}/lib/charset.alias"))

        # Main binary
        path = self.project.get_main_binary()
        source = self.project.evaluate_path(path.source)
        if not os.path.exists(source):
            print "Cannot find main binary: " + source
            sys.exit(1)

        dest = self.copy_path(path)
        self.binary_paths.append(dest)

        # Additional binaries (executables, libraries, modules)
        self.copy_binaries(self.project.get_binaries())
        self.resolve_library_dependencies()

        # Data
        for path in self.project.get_data():
            self.copy_path(path)
        
        # Translations
        self.copy_translations()

        # Frameworks
        frameworks = self.project.get_frameworks()
        for path in frameworks:
            dest = self.copy_path(path)
            self.frameworks.append(dest)

        self.copy_icon_themes()

        self.create_pango_setup()
        self.create_gtk_immodules_setup()
        self.create_gdk_pixbuf_loaders_setup()

        if self.meta.run_install_name_tool:
            self.run_install_name_tool()

        #self.strip_debugging()

        self.sign_binaries()

        # Launcher script, if necessary.
        launcher_script = self.project.get_launcher_script()
        if launcher_script:
            path = self.copy_path(launcher_script)
            if os.environ.has_key("APPLICATION_CERT"):
                cert = os.environ["APPLICATION_CERT"]
                ident = self.project.get_bundle_id()
                cmdargs = ['codesign', '-s', cert, '-i', ident, "-f", path]
                result = os.spawnvp(os.P_WAIT, 'codesign', cmdargs)
                if result:
                    raise OSError, '"'+ " ".join(cmdargs) + '" failed %d' % result
        if self.meta.overwrite:
            self.recursive_rm(final_path)
        shutil.move(self.project.get_bundle_path(), final_path)

if __name__ == '__main__':
    if len(sys.argv) != 2:
        print "Usage: %s <bundle descriptopn file>" % (sys.argv[0])
        sys.exit(2)

    if not os.path.exists(sys.argv[1]):
        print "File %s does not exist" % (sys.argv[1])
        sys.exit(2)

    project = Project(sys.argv[1])
    bundler = Bundler(project)

    bundler.run()



#!/usr/bin/python
import unittest
import os
from project_test import Project_Test

def setProjects( goodpath, badpath):
    if not os.path.isabs(goodpath):
        goodpath = os.path.join(os.getcwd(), goodpath)
    f = open(goodpath)
    Project_Test.goodxml = f.read()
    f.close()
    Project_Test.goodpath = goodpath
    if not os.path.isabs(badpath):
        badpath = os.path.join(os.getcwd(), badpath)
    f = open(badpath)
    Project_Test.badxml = f.read()
    f.close()
    Project_Test.badpath = badpath
 
setProjects("test/goodproject.bundle", "test/badproject.bundle")
suite = unittest.TestLoader().loadTestsFromTestCase(Project_Test)
unittest.TextTestRunner(verbosity=2).run(suite)



import sys, os

from project import *
from bundler import *

def main(argv):
    if len(argv) != 1:
        print "Usage: %s <bundle descriptopn file>" % (sys.argv[0])
        sys.exit(2)

    if not os.path.exists(argv[0]):
        print "File %s does not exist" % (argv[0])
        sys.exit(2)

    project = Project(argv[0])
    bundler = Bundler(project)

    bundler.run()



import re
import os
import errno

def evaluate_environment_variables(string):
    p = re.compile("\${env:(.+?)}")
    m = p.search(string)
    while m:
        env = m.group(1)
        value = os.getenv(env)
        if not value:
            raise Exception("Environment variable %s is undefined" % (env))
        string = p.sub(value, string, 1)
        m = p.search(string)

    return string

def evaluate_pkgconfig_variables(string):
    p = re.compile("\${pkg:(.*?):(.*?)}")
    m = p.search(string)
    while m:
        module = m.group(1)
        key = m.group(2)
        f = os.popen("pkg-config --variable=" + key + " " + module)
        value = f.read().strip()
        if not value:
            raise Exception("pkg-config variable '%s %s' is undefined" % (key, module))
        string = p.sub(value, string, 1)
        m = p.search(string)

    return string

def makedirs(path):
    try:
        os.makedirs(path)
    except EnvironmentError, e:
        if e.errno != errno.EEXIST:
            raise

def node_get_elements_by_tag_name(node, name):
    try:
        return node.getElementsByTagName(name)
    except:
        return []

def node_get_element_by_tag_name(node, name):
    try:
        (data,) = node.getElementsByTagName(name)
        return data 
    except:
        return None

def node_get_string(node, default=None):
    try:
        return node.firstChild.data.strip()
    except:
        return default

def node_get_property_boolean(node, name, default=False):
    try:
        value = node.getAttribute(name)
        if value in [ "true", "yes" ]:
            return True
    except:
        pass

    return default

def filterlines(p, lines):
    for line in lines:
        match = p.match(line)
        if match:
            yield match.group(1)



import os
import errno
import sys
import unittest
import xml.dom.minidom
from xml.dom.minidom import Node
from plistlib import Plist
from project import Project
import utils

class Mock_Project(Project):

    def __init__(self, testxml, path):
        doc = xml.dom.minidom.parseString(testxml)
        self.root = utils.node_get_element_by_tag_name(doc, "app-bundle")
        assert self.root != None
        dir, tail = os.path.split(path)
        self.project_dir = os.path.join(os.getcwd(), dir)
        self.project_path = os.path.join(self.project_dir, tail)
        try:
            plist_path = os.path.join(self.project_dir, "test.plist")
            plist = Plist.fromFile(plist_path)
        except EnvironmentError, e:
            if e.errno == errno.ENOENT:
                print "Info.plist file not found: " + plist_path
                sys.exit(1)
            else:
                raise
        self.name = plist.CFBundleExecutable

class Project_Test(unittest.TestCase):

    def setUp(self):
        self.goodproject = Mock_Project(Project_Test.goodxml, 
                                        Project_Test.goodpath)
        self.badproject = Mock_Project(Project_Test.badxml,
                                       Project_Test.badpath)

    def test_a_get_project_path(self):
        path = self.goodproject.get_project_path()
        self.failUnlessEqual(path, Project_Test.goodpath,
                   "Project returned incorrect project path %s wanted %s" % 
                             (path, Project_Test.goodpath) )
        path = self.badproject.get_project_path()
        self.failUnlessEqual(path, Project_Test.badpath,
                   "Project returned incorrect bad project path %s" % path)

    def test_b_get_project_dir(self):
        dir = self.goodproject.get_project_dir()
        good_dir, tail = os.path.split(Project_Test.goodpath)
        self.failUnlessEqual(dir, good_dir,
                   "Project returned incorrect project dir %s" % dir)

    def test_c_get_meta(self):
        node = self.goodproject.get_meta()
        self.failIfEqual(node, None, "Didn't find meta node in goodproject")
        self.assertRaises(Exception, self.badproject.get_meta())

    def test_d_get_prefix(self):
        try:
            pfx = self.goodproject.get_prefix()
        except KeyError:
            self.fail("Goodproject didn't set the default prefix")
        self.failUnlessEqual(pfx, os.getenv("JHBUILD_PREFIX", 
                                            "Default prefix %s" % pfx))
        try:
            pfx = self.goodproject.get_prefix("alt")
        except KeyError:
            self.fail("Goodproject didn't set the alt prefix")
        self.failUnlessEqual(pfx, "/usr/local/gtk", "Alternate prefix %s" % pfx)

    def test_e_get_plist_path(self):
        try:
            path = self.goodproject.get_plist_path()
        except KeyError:
            self.fail("Goodproject didn't set the default prefix")
        except Exception:
            self.fail("Goodproject didn't set the plist tag")
        dir, tail = os.path.split(Project_Test.goodpath)
        self.failUnlessEqual(path, 
                             os.path.join(dir, "test.plist"),
                             "Bad Plist Path %s" % path)
        self.failUnlessRaises(Exception, self.badproject.get_plist_path)

    def test_f_get_name(self):
        try:
            plist_path = self.goodproject.get_plist_path()
        except KeyError:
            self.fail("Goodproject didn't set the default prefix")
        try:
            plist = Plist.fromFile(plist_path)
            name = plist.CFBundleExecutable
        except IOError:
            self.fail("Path problem " + plist_path)
        pname = self.goodproject.get_name()
        self.failUnlessEqual(pname, name, "Bad Name %s" % pname)



#If get_name works, this will too.
    def test_g_get_bundle_path(self):
        pass 

    def test_h_evaluate_path(self):
        try:
            path = self.goodproject.evaluate_path("${bundle}/Contents/MacOS/${name}")
        except KeyError:
            self.fail("Goodproject didn't set the default prefix")
        try:
            name = self.goodproject.get_bundle_path()
        except AttributeError:
            self.fail("Project didn't set the name attribute")
        self.failUnlessEqual(path, os.path.join(name, os.path.join("Contents/MacOS", self.goodproject.name)), "Bundle path evaluation failed %s" % path)
        path = self.goodproject.evaluate_path("${prefix}/bin/foo")
        self.failUnlessEqual(path, 
                             os.path.join(self.goodproject.get_prefix(), 
                                          "bin/foo"), 
                             "Prefix path evaluation failed %s" % path)

    def test_i_get_launcher_script(self):
        path = self.goodproject.get_launcher_script()
        self.failUnlessEqual(path.source, 
                    os.path.join(os.path.dirname(__file__), "launcher.sh"),
                    "Bad launcher source")
        self.failUnlessEqual(path.dest, "${bundle}/Contents/MacOS/${name}",
                    "Bad launcher destination")

    def test_j_get_icon_themes(self):
        themes = self.goodproject.get_icon_themes()
        self.failUnlessEqual(len(themes), 2, 
                             "Wrong number of themes %d" % len(themes))
        self.failUnlessEqual(themes[-1].name, "hicolor", 
                             "No hicolor theme %s" % themes[-1].name)

    def test_k_get_environment(self):
        env = self.goodproject.get_environment()
        self.failUnlessEqual(len(env.runtime_variables), 2, 
                         "Wrong number of runtime variables %d" % 
                             len(env.runtime_variables))
        self.failUnlessEqual(len(env.scripts), 2, 
                             "Wrong number of scripts %d" % len(env.scripts))

    def test_l_get_frameworks(self):
        fw = self.goodproject.get_frameworks()
        self.failUnlessEqual(len(fw), 1, 
                             "Wrong number of frameworks %d" % len(fw))

    def test_m_get_main_binary(self):
        bin = self.goodproject.get_main_binary()
        self.failUnlessEqual(bin.source, "${prefix}/bin/foo-source", 
                         "Bad binary source %s" % bin.source)
        self.failUnlessEqual(bin.dest, "${bundle}/Contents/MacOS/${name}-bin", 
                    "Bad binary destination %s" % bin.dest)

    def test_n_get_binaries(self):
        bin = self.goodproject.get_binaries()
        self.failUnlessEqual(len(bin), 2, 
                             "Wrong number of binaries %d" % len(bin))

    def test_o_get_data(self):
        data = self.goodproject.get_data()
        self.failUnlessEqual(len(data), 3, 
                             "Wrong number of data paths %d" % len(data))
        self.failUnlessEqual(data[2].dest, 
                             "${bundle}/Contents/Resources/etc/gtk-2.0/gtkrc", 
                             "Data[2] Destination %s" % data[2].dest)

    def test_p_get_translations(self):
        trans = self.goodproject.get_translations()
        self.failUnlessEqual(len(trans), 1, 
                             "Wrong number of translations %d" % len(trans))
        self.failUnlessEqual(trans[0].name, "foo", 
                             "Bad translation name %s" % trans[0].name)
        self.failUnlessEqual(trans[0].source, "${prefix}/share/locale",
                             "Bad translation source %s" % trans[0].source)


        



import errno
import sys
import re
import os
import xml.dom.minidom
from xml.dom.minidom import Node
from plistlib import Plist
import utils

# Base class for anything that can be copied into a bundle with a
# source and dest.
class Path:
    def __init__(self, source, dest=None, recurse=False):
        if source and len(source) == 0:
            source = None
        if dest and len(dest) == 0:
            dest = None

        if source and os.path.isabs(source):
            source = os.path.normpath(source)
        if dest and os.path.isabs(dest):
            dest = os.path.normpath(dest)
            
        self.source = source
        self.dest = dest
        self.recurse = recurse

    def from_node(cls, node, validate=True):
        source = utils.node_get_string(node)
        dest = node.getAttribute("dest")
        recurse = node.getAttribute("recurse")
        if len(dest) == 0:
            dest = None
        if recurse == "True":
            recurse = True
        else:
            recurse = False
        if validate:
            Path.validate(source, dest)
            
        return Path(source, dest, recurse)
    from_node = classmethod(from_node)

    def validate(cls, source, dest):
        if source and len(source) == 0:
            source = None
        if dest and len(dest) == 0:
            dest = None

        if not source or len(source) == 0:
            raise Exception("The source path cannot be empty")
        
        if source.startswith("${bundle}"):
            raise Exception("The source path cannot use a ${bundle} macro")

        if dest and dest.startswith("${prefix"):
            raise Exception("The destination path cannot use a ${prefix} macro")

        if not os.path.isabs(source):
            if not (source.startswith("${project}") or source.startswith("${env:") or \
                    source.startswith("${pkg:") or source.startswith("${prefix}") or \
                    source.startswith("${prefix:")):
                raise Exception("The source path must be absolute or use one of the "
                                "predefined macros ${project}, ${prefix}, ${prefix:*}, "
                                "${env:*}, or ${pkg:*:*}")

        if not source.startswith("${prefix") and os.path.isabs(source):
            if not dest:
                raise Exception("If the source doesn't use a ${prefix} or ${prefix:*} "
                                "macro, the destination path must be set " + dest)

        if not dest and not source.startswith("${prefix"):
            raise Exception("If the destination path is empty, the source must use "
                            "a ${prefix} or ${prefix:*} macro")

        if dest and len(dest) > 0 and not dest.startswith("${bundle}"):
            raise Exception("The destination path must start with ${bundle}")

        return True
    validate = classmethod(validate)

# Used for anything that has a name and value.
class Variable:
    def __init__(self, node):
        self.name = node.getAttribute("name")
        self.value = utils.node_get_string(node)

class Environment:
    def __init__(self, node):
        self.runtime_variables = []
        self.scripts = []

        variables = utils.node_get_elements_by_tag_name(node, "runtime-variable")
        for child in variables:
            self.runtime_variables.append(Variable(child))

        scripts = utils.node_get_elements_by_tag_name(node, "script")
        for child in scripts:
            script = Path(utils.node_get_string(child), "${bundle}/Resources/Scripts")
            self.scripts.append(script)

class Meta:
    def __init__(self, node):
        self.prefixes = {}

        prefixes = utils.node_get_elements_by_tag_name(node, "prefix")
        for child in prefixes:
            name = child.getAttribute("name")
            if len(name) == 0:
                name = "default"
            value = utils.evaluate_environment_variables(utils.node_get_string(child))
            self.prefixes[name] = value

        child = utils.node_get_element_by_tag_name(node, "image")
        if child:
            pass # FIXME: implement
        
        child = utils.node_get_element_by_tag_name(node, "run-install-name-tool")
        if child:
            self.run_install_name_tool = True
        else:
            self.run_install_name_tool = False

        child = utils.node_get_element_by_tag_name(node, "destination")
        self.overwrite = utils.node_get_property_boolean(child, "overwrite", False)
        self.dest = utils.node_get_string(child, "${project}")

        child = utils.node_get_element_by_tag_name(node, "gtk")
        if child:
            self.gtk = utils.node_get_string(child)
        else:
            self.gtk = "gtk+-2.0"

class Framework(Path):
    def __init__(self, source):
        Path.__init__(self, source, self.get_dest_path_from_source(source))

    def from_node(cls, node):
        framework = Path.from_node(node, validate=False)
        framework.dest = Framework.get_dest_path_from_source(framework.source)

        return framework
    from_node = classmethod(from_node)

    def get_dest_path_from_source(cls, source):
        (head, tail) = os.path.split(source)
        return "${bundle}/Contents/Frameworks/" + tail
    get_dest_path_from_source = classmethod(get_dest_path_from_source)

class Binary(Path):
    def __init__(self, source, dest):
        Path.__init__(self, source, dest)

    def from_node(cls, node):
        binary = Path.from_node(node)

        if not binary.source:
            raise "The tag 'binary' must have a 'source' property"

        return binary
    from_node = classmethod(from_node)

class Translation(Path):
    def __init__(self, name, sourcepath, destpath):
        Path.__init__(self, sourcepath, destpath)
        self.name = name

    def from_node(cls, node):
        source = utils.node_get_string(node)
        dest = node.getAttribute("dest")
        if len(dest) == 0:
            dest = None
        name = node.getAttribute("name")
        if len(name) == 0:
            raise "The tag 'translations' must have a 'name' property"

        return Translation(name, source, dest)
    from_node = classmethod(from_node)


        

class Data(Path):
    pass

class IconTheme:
    ICONS_NONE, ICONS_ALL, ICONS_AUTO = range(3)
    
    def __init__(self, name, icons=ICONS_AUTO):
        self.name = name
        self.source = "${prefix}/share/icons/" + name

        self.icons = icons

    def from_node(cls, node):
        name = utils.node_get_string(node)
        if not name:
            raise Exception("Icon theme must have a 'name' property")
        
        string = node.getAttribute("icons")
        if string == "all":
            icons = IconTheme.ICONS_ALL
        elif string == "none":
            icons = IconTheme.ICONS_NONE
        elif string == "auto" or len(string) == 0:
            icons = IconTheme.ICONS_AUTO
            
        return IconTheme(name, icons)
    from_node = classmethod(from_node)

class Project:
    def __init__(self, project_path=None):
        if not os.path.isabs(project_path):
            project_path = os.path.join(os.getcwd(), project_path)
        self.project_path = project_path
        self.root = None

        if project_path and os.path.exists(project_path):
            try:
                doc = xml.dom.minidom.parse(project_path)
                # Get the first app-bundle tag and ignore any others.
                self.root = utils.node_get_element_by_tag_name(doc, "app-bundle")
            except:
                print "Could not load project %s:" % (project_path)
                raise

        # The directory the project file is in (as opposed to
        # project_path which is the path including the filename).
        self.project_dir, tail = os.path.split(project_path)
        self.meta = self.get_meta()

        plist_path = self.get_plist_path()
        try:
            plist = Plist.fromFile(plist_path)
        except EnvironmentError, e:
            if e.errno == errno.ENOENT:
                print "Info.plist file not found: " + plist_path
                sys.exit(1)
            else:
                raise
        self.name = plist.CFBundleExecutable
        if plist.has_key("CFBundleName"):
            self.bundle_name = plist.CFBundleName
        else:
            self.bundle_name = plist.CFBundleExecutable

        self.bundle_id = plist.CFBundleIdentifier

    """
     Replace ${env:?}, ${prefix}, ${prefix:?}, ${project}, ${gtk}, ${gtkdir},
     ${gtkversion}, ${pkg:?:?}, ${bundle}, and ${name} variables.
    """
    def evaluate_path(self, path, include_bundle=True):
        p = re.compile("^\${prefix}")
        path = p.sub(self.get_prefix(), path)

        p = re.compile("^\${prefix:(.*?)}")
        m = p.match(path)
        if m:
            path = p.sub(self.get_prefix(m.group(1)), path)

        p = re.compile("^\${project}")
        path = p.sub(self.project_dir, path)

        p = re.compile("\${gtk}")
        path = p.sub(self.meta.gtk, path)

        p = re.compile("\${gtkdir}")
        path = p.sub(self.get_gtk_dir(), path)

        p = re.compile("\${gtkversion}")
        path = p.sub(self.get_gtk_version(), path)

        try:
            p = re.compile("\${name}")
            path = p.sub(self.name, path)
        except AttributeError:
            pass # can be used before name path is set

        if include_bundle:
            try:
                p = re.compile("^\${bundle}")
                path = p.sub(self.get_bundle_path(), path)
            except AttributeError:
                pass # can be used before bundle path is set

        path = utils.evaluate_environment_variables(path)
        path = utils.evaluate_pkgconfig_variables(path)

        return os.path.normpath(path)

    def get_name(self):
        return self.name

    def get_bundle_name(self):
        return self.bundle_name

    def get_bundle_id(self):
        return self.bundle_id

    def get_prefix(self, name="default"):
        meta = self.get_meta()
        return meta.prefixes[name]

    def get_project_path(self):
        return self.project_path

    def get_project_dir(self):
        return self.project_dir

    def get_bundle_path(self, *args):
        dest = self.get_meta().dest
        bundle_path = os.path.join(dest, "." + self.get_name() + ".app")
        bundle_path = self.evaluate_path(bundle_path, False)
        return os.path.join(bundle_path, *args)

    def get_plist_path(self):
        plist = utils.node_get_element_by_tag_name(self.root, "plist")
        if not plist:
            raise Exception("The 'plist' tag is required")
        return  self.evaluate_path(utils.node_get_string(plist))

    def get_launcher_script(self):
        node = utils.node_get_element_by_tag_name(self.root, "launcher-script")
        if node:
            path = Path.from_node(node, False)
            path.dest = "${bundle}/Contents/MacOS/${name}"
        else:
            # Use the default launcher.
            launcher = os.path.join(os.path.dirname(__file__), "launcher.sh")
            path = Path(launcher, "${bundle}/Contents/MacOS/${name}")

        return path

    def get_icon_themes(self):
        themes = []

        nodes = utils.node_get_elements_by_tag_name(self.root, "icon-theme")
        for node in nodes:
            themes.append(IconTheme.from_node(node))

        # The hicolor theme is mandatory.
        if not filter(lambda l: l.name == "hicolor", themes):
            themes.append(IconTheme("hicolor"))

        return themes

    def get_meta(self):
        node = utils.node_get_element_by_tag_name(self.root, "meta")
        return Meta(node)

    def get_gtk_version(self):
        if self.meta.gtk == "gtk+-3.0":
            return "3.0"
        else:
            return "2.0"

    def get_gtk_dir(self):
        if self.meta.gtk == "gtk+-3.0":
            return "gtk-3.0"
        else:
            return "gtk-2.0"

    def get_environment(self):
        node = utils.node_get_element_by_tag_name(self.root, "environment")
        return Environment(node)

    def get_frameworks(self):
        frameworks = []
        nodes = utils.node_get_elements_by_tag_name(self.root, "framework")
        for node in nodes:
            frameworks.append(Framework.from_node(node))
        return frameworks

    def get_translations(self):
        translations = []
        nodes = utils.node_get_elements_by_tag_name(self.root, "translations")
        for node in nodes:
            translations.append(Translation.from_node(node))
        return translations

    def get_main_binary(self):
        node = utils.node_get_element_by_tag_name(self.root, "main-binary")
        if not node:
            raise Exception("The file has no <main-binary> tag")

        binary = Binary.from_node(node)

        launcher = self.get_launcher_script()
        if launcher:
            suffix = "-bin"
        else:
            suffix = ""
        binary.dest = "${bundle}/Contents/MacOS/${name}" + suffix

        return binary

    def get_binaries(self):
        binaries = []
        nodes = utils.node_get_elements_by_tag_name(self.root, "binary")
        for node in nodes:
            binaries.append(Binary.from_node(node))
        return binaries

    def get_data(self):
        data = []
        nodes = utils.node_get_elements_by_tag_name(self.root, "data")
        for node in nodes:
            data.append(Data.from_node(node))
        return data
            
if __name__ == '__main__':
    project = Project(os.path.join(os.getcwd(), 'giggle.bundle'))

    print "General:"
    print "  Project path: %s" % (project.get_project_path())
    print "  Plist path: %s" % (project.get_plist_path())
    print "  App name: %s" % (project.name)
    print "  Destination: %s" % (project.get_meta().dest)
    print "  Overwrite: %s" % (str(project.get_meta().overwrite))

    environment = project.get_environment()
    print "Environment:"
    for variable in environment.runtime_variables:
        print "  %s=%s" % (variable.name, variable.value)
    for script in environment.scripts:
        print "  %s => %s" % (script.source, script.dest)

    print "Frameworks:"
    for framework in project.get_frameworks():
        print " ", framework

    print "Main binary:"
    binary = project.get_main_binary()
    print "  %s => %s" % (binary.source, binary.dest)

    print "Launcher:"
    launcher_script = project.get_launcher_script()
    print "  %s => %s" % (launcher_script.source, launcher_script.dest)

    print "Binaries:"
    for binary in project.get_binaries():
        print "  %s => %s" % (binary.source, binary.dest)

